/*

        Decode file as MIME base64 (RFC 1341)
                  by John Walker
              http://www.fourmilab.ch/

        This program is in the public domain.

Adapted by Fredrik Rambris to decode into the same string
in memory instead of from stdin to stdout.

*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#define TRUE  1
#define FALSE 0

/*  INSIG  --  Return next significant input  */
static char insig( char **p )
{
	char c;

	/*CONSTANTCONDITION*/
	while( TRUE )
	{
		c = ( *p )[0];
		( *p )++;
		if( c == '\0' || c > ' ' ) return ( c );
	}
 /*NOTREACHED*/}

/*  OUTNCHARS  --  Output n chars from c to p.  */
void outnchars( char **p, char *c, int n )
{
	if( !n ) return;
	while( n-- )
	{
		( *p )[0] = c[0];
		c++;
		( *p )++;
	}
	( *p )[0] = '\0';
}

/*  DECODE  --	Decode base64.	*/
void base64_decode( char *str )
{
	int i;
	char *in = str, *out = str;
	char dtable[256];

	if( !str ) return;
	for( i = 0; i < 255; i++ ) dtable[i] = 0x80;

	for( i = 'A'; i <= 'Z'; i++ ) dtable[i] = 0 + ( i - 'A' );

	for( i = 'a'; i <= 'z'; i++ ) dtable[i] = 26 + ( i - 'a' );

	for( i = '0'; i <= '9'; i++ ) dtable[i] = 52 + ( i - '0' );

	dtable['+'] = 62;
	dtable['/'] = 63;
	dtable['='] = 0;

	/*CONSTANTCONDITION*/
	while( TRUE )
	{
		char a[4], b[4], o[3];

		for( i = 0; i < 4; i++ )
		{
			int c = insig( &in );

			if( c == '\0' ) return;

			if( dtable[c] & 0x80 )
			{
				/* Ignoring errors: discard invalid character. */
				i--;
				continue;
			}
			a[i] = ( char ) c;
			b[i] = ( char ) dtable[c];
		}
		o[0] = ( b[0] << 2 ) | ( b[1] >> 4 );
		o[1] = ( b[1] << 4 ) | ( b[2] >> 2 );
		o[2] = ( b[2] << 6 ) | b[3];
		i = a[2] == '=' ? 1 : ( a[3] == '=' ? 2 : 3 );

		outnchars( &out, o, i );

		if( i < 3 ) return;
	}
}

