/*
    XMMSd - A web control plug-in for XMMS
    Copyright (C) 2002  Fredrik Rambris <boost@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* plugin.c - Code for starting the server in plug-in mode */

#include <config.h>
#include "types.h"
#include "server.h"
#include "response.h"
#include <gtk/gtk.h>
#include <plugin.h>
#include <configfile.h>

static void init(void);
static void cleanup(void);
static void about(void);
static void configure( void );

void restart_server( void );
void read_config( void );
void write_config( void );
gboolean delete_event( GtkWidget *widget, GdkEvent *event, gpointer data );
void destroy (GtkWidget *widget, gpointer data);
void apply_clicked( GtkButton *button, gpointer user_data );
void cancel_clicked( GtkButton *button, gpointer user_data );
void ok_clicked( GtkButton *button, gpointer user_data );
void about_ok_clicked( GtkButton *button, gpointer user_data );
void require_login_toggled( GtkToggleButton *togglebutton, gpointer user_data );
void xc_free( void );

static struct Server *serv=NULL;
static gint port=8335;
static gchar *bind_to=NULL;
static gchar cport[6]="";
static struct xmmsd_config *xc=NULL;

static GtkWidget *configure_win=NULL, *bind_to_entry, *port_entry, *about_win=NULL;
static GtkWidget *require_login_checkbox, *user_entry, *password_entry, *user_label, *password_label;

GeneralPlugin xmmsd_gp =
{
	NULL,			/* handle */
	NULL,			/* filename */
	-1,			/* xmms_session */
	NULL,			/* Description */
	init,
	about,
	configure,
	cleanup,
};

GeneralPlugin *get_gplugin_info(void)
{
	xmmsd_gp.description = g_strdup_printf("XMMSd web control %s", VERSION);
	return &xmmsd_gp;
}

static void init(void)
{
	read_config();
	restart_server();
}

static void cleanup(void)
{
	server_free( serv );
	serv=NULL;
	xc_free();
}

static void about(void)
{
	GtkWidget *vbox, *title_label, *mez_frame, *mez_vbox, *mez_label;
	GtkWidget *ok;

	if( about_win!=NULL) return;

	about_win=gtk_window_new( GTK_WINDOW_DIALOG );
	gtk_signal_connect( GTK_OBJECT( about_win ), "destroy", GTK_SIGNAL_FUNC( gtk_widget_destroyed ), &about_win );
	gtk_signal_connect( GTK_OBJECT( about_win ), "delete_event", GTK_SIGNAL_FUNC(delete_event), NULL );
	gtk_window_set_title( GTK_WINDOW( about_win ), "About XMMSd");
	gtk_container_set_border_width( GTK_CONTAINER( about_win ), 8 );
	gtk_window_set_policy( GTK_WINDOW( about_win ), FALSE, FALSE, TRUE );
	
		vbox=gtk_vbox_new( FALSE, 4 );
		gtk_container_add( GTK_CONTAINER( about_win ), vbox );

			mez_frame=gtk_frame_new( NULL );
			gtk_frame_set_shadow_type( GTK_FRAME( mez_frame ), GTK_SHADOW_IN );
			gtk_box_pack_start( GTK_BOX( vbox ), mez_frame, TRUE, TRUE, 2 );
			
				mez_vbox=gtk_vbox_new( FALSE, 4 );
				gtk_container_set_border_width( GTK_CONTAINER( mez_vbox ), 8 );
				gtk_container_add( GTK_CONTAINER( mez_frame ), mez_vbox );

					title_label=gtk_label_new( "XMMSd web control plug-in v" VERSION );
					gtk_box_pack_start( GTK_BOX( mez_vbox ), title_label, FALSE, FALSE, 2 );

					mez_label=gtk_label_new( "Created by Fredrik Rambris <boost@users.sourceforge.net>\nhttp://xmmsd.sourceforge.net/\n\nControl XMMS from a web browser\n\nXMMSd comes with ABSOLUTELY NO WARRANTY.\nThis is free software, and you are welcome\nto redistribute it under terms of the\nGNU General Public License.\nSee COPYING file for more information." );
					gtk_box_pack_start( GTK_BOX( mez_vbox ), mez_label, FALSE, FALSE, 2 );

					gtk_widget_show( mez_label );
					gtk_widget_show( title_label );
				gtk_widget_show( mez_vbox );

			ok=gtk_button_new_with_label( "Ok" );
			gtk_signal_connect( GTK_OBJECT( ok ), "clicked", GTK_SIGNAL_FUNC(about_ok_clicked), NULL );
			gtk_box_pack_start( GTK_BOX( vbox ), ok, FALSE, FALSE, 2 );

			gtk_widget_show( ok );
			gtk_widget_show( mez_frame );
		gtk_widget_show( vbox );
	gtk_widget_show( about_win );
}

static void configure( void )
{
	GtkWidget *main_vbox, *notebook, *button_hbox, *network_table, *network_label, *authentication_table, *authentication_label;
	GtkWidget *ok_button, *cancel_button, *apply_button;
	GtkWidget *bind_to_label, *port_label;
	GtkWidget *network_info_label;
	GtkWidget *require_login_label;
	GtkTooltips *tooltips;

	if( configure_win ) return;
	if( !bind_to ) read_config();
	sprintf( cport, "%d", port );
	
	tooltips=gtk_tooltips_new();
	configure_win = gtk_window_new( GTK_WINDOW_DIALOG );
	gtk_signal_connect( GTK_OBJECT(configure_win), "delete_event", GTK_SIGNAL_FUNC(delete_event), NULL );
	gtk_signal_connect( GTK_OBJECT(configure_win), "destroy", GTK_SIGNAL_FUNC(destroy), NULL );
	gtk_window_set_title( GTK_WINDOW(configure_win), "XMMSd Configuration");
	gtk_container_set_border_width( GTK_CONTAINER( configure_win ), 2 );
	gtk_window_set_policy( GTK_WINDOW( configure_win ), FALSE, FALSE, TRUE );

		main_vbox=gtk_vbox_new( FALSE, 4 );
		gtk_container_add( GTK_CONTAINER( configure_win ), main_vbox );

			notebook=gtk_notebook_new();
			gtk_box_pack_start( GTK_BOX( main_vbox ), notebook, FALSE, FALSE, 0 );

				network_table=gtk_table_new( 2, 3, FALSE );
				gtk_container_set_border_width( GTK_CONTAINER( network_table ), 2 );
				network_label=gtk_label_new( "Network" );
				gtk_notebook_append_page( GTK_NOTEBOOK( notebook ), network_table, network_label );

					bind_to_label=gtk_label_new( "Bind to" );
					gtk_table_attach( GTK_TABLE( network_table ), bind_to_label, 0, 1, 0, 1, GTK_SHRINK, GTK_SHRINK, 0, 0 );

					bind_to_entry=gtk_entry_new_with_max_length(15);
					if( bind_to ) gtk_entry_set_text( GTK_ENTRY( bind_to_entry ), bind_to );
					gtk_tooltips_set_tip( tooltips, bind_to_entry, "IP-address of a local interface on which to allow incomming connection or * (asterisk) for all interfaces", "" );
					gtk_table_attach( GTK_TABLE( network_table ), bind_to_entry, 1, 2, 0, 1, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 0 );

					port_label=gtk_label_new( "Port" );
					gtk_table_attach( GTK_TABLE( network_table ), port_label, 0, 1, 1, 2, GTK_SHRINK, GTK_SHRINK, 0, 0 );

					port_entry=gtk_entry_new_with_max_length(5);
					gtk_entry_set_text( GTK_ENTRY( port_entry ), cport );
					gtk_table_attach( GTK_TABLE( network_table ), port_entry, 1, 2, 1, 2, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 0 );

					network_info_label=gtk_label_new( "NOTE! When changing any of these options the server is restarted and may be unavailable for a short while" );
					gtk_label_set_justify( GTK_LABEL( network_info_label ), GTK_JUSTIFY_LEFT );
					gtk_label_set_line_wrap( GTK_LABEL( network_info_label ), TRUE );
					gtk_table_attach( GTK_TABLE( network_table ), network_info_label, 0, 2, 2, 3, GTK_SHRINK, GTK_SHRINK, 0, 0 );

					gtk_widget_show( network_info_label );
					gtk_widget_show( port_entry );
					gtk_widget_show( port_label );
					gtk_widget_show( bind_to_entry );
					gtk_widget_show( bind_to_label );

				authentication_table=gtk_table_new( 2,4, FALSE );
				gtk_container_set_border_width( GTK_CONTAINER( authentication_table ), 2 );
				authentication_label=gtk_label_new( "Authentication" );
				gtk_notebook_append_page( GTK_NOTEBOOK( notebook ), authentication_table, authentication_label );

					require_login_label=gtk_label_new( "Require login" );
					gtk_table_attach( GTK_TABLE( authentication_table ), require_login_label, 0, 1, 0, 1, GTK_SHRINK, GTK_SHRINK, 0, 0 );
					require_login_checkbox=gtk_check_button_new();
					gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( require_login_checkbox ), xc->require_login );
					gtk_signal_connect( GTK_OBJECT(require_login_checkbox), "toggled", GTK_SIGNAL_FUNC(require_login_toggled), NULL );
					gtk_table_attach( GTK_TABLE( authentication_table ), require_login_checkbox, 1, 2, 0, 1, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 0 );


					user_label=gtk_label_new( "User name" );
					gtk_table_attach( GTK_TABLE( authentication_table ), user_label, 0, 1, 1, 2, GTK_SHRINK, GTK_SHRINK, 0, 0 );
					user_entry=gtk_entry_new();
					if( xc->user && strlen( xc->user ) ) gtk_entry_set_text( GTK_ENTRY( user_entry ), xc->user );
					else gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( require_login_checkbox ), FALSE );
					gtk_table_attach( GTK_TABLE( authentication_table ), user_entry, 1, 2, 1, 2, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 0 );
					password_label=gtk_label_new( "Password" );
					gtk_table_attach( GTK_TABLE( authentication_table ), password_label, 0, 1, 2, 3, GTK_SHRINK, GTK_SHRINK, 0, 0 );
					password_entry=gtk_entry_new();
					if( xc->password && strlen( xc->password ) ) gtk_entry_set_text( GTK_ENTRY( password_entry ), xc->password );
					else gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( require_login_checkbox ), FALSE );
					/* If we don't have any username or password we don't want login to be required */
					require_login_toggled( GTK_TOGGLE_BUTTON( require_login_checkbox ), NULL );
					gtk_table_attach( GTK_TABLE( authentication_table ), password_entry, 1, 2, 2, 3, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 0 );

					gtk_widget_show( user_label );
					gtk_widget_show( user_entry );
					gtk_widget_show( password_label );
					gtk_widget_show( password_entry );
					gtk_widget_show( require_login_checkbox );
					gtk_widget_show( require_login_label );
				
				gtk_widget_show( authentication_label );
				gtk_widget_show( authentication_table );
				gtk_widget_show( network_label );
				gtk_widget_show( network_table );

			button_hbox=gtk_hbox_new( TRUE, 4 );
			gtk_box_pack_start( GTK_BOX( main_vbox ), button_hbox, FALSE, FALSE, 0 );

				ok_button=gtk_button_new_with_label( "Ok" );
				gtk_signal_connect( GTK_OBJECT(ok_button), "clicked", GTK_SIGNAL_FUNC(ok_clicked), NULL );
				gtk_box_pack_start( GTK_BOX( button_hbox ), ok_button, FALSE, TRUE, 2 );

				cancel_button=gtk_button_new_with_label( "Cancel" );
				gtk_signal_connect( GTK_OBJECT(cancel_button), "clicked", GTK_SIGNAL_FUNC(cancel_clicked), NULL );
				gtk_box_pack_start( GTK_BOX( button_hbox ), cancel_button, FALSE, TRUE, 2 );

				apply_button=gtk_button_new_with_label( "Apply" );
				gtk_signal_connect( GTK_OBJECT(apply_button), "clicked", GTK_SIGNAL_FUNC(apply_clicked), NULL );
				gtk_box_pack_start( GTK_BOX( button_hbox ), apply_button, FALSE, TRUE, 2 );

				gtk_widget_show( apply_button );
				gtk_widget_show( cancel_button );
				gtk_widget_show( ok_button );

			gtk_widget_show( button_hbox );
			gtk_widget_show( notebook );
		gtk_widget_show( main_vbox );
	gtk_widget_show( configure_win );
}

void restart_server( void )
{
	server_free( serv );
	serv=server_new(
			SERVER_BIND, GPOINTER_TO_INT( bind_to ),
			SERVER_PORT, port,
			SERVER_RESPONSE, GPOINTER_TO_INT( handle_http_response ),
			SERVER_AUTHENTICATION, GPOINTER_TO_INT( auth_handler ),
			XMMSD_SESSION, xmmsd_gp.xmms_session,
			SERVER_USERDATA, GPOINTER_TO_INT( xc ),
			TAG_DONE );
}

void read_config( void )
{
	ConfigFile *cfgfile;
	gchar *bt;

	if( !xc ) xc=g_new0( struct xmmsd_config, 1 );

	if( ( cfgfile=xmms_cfg_open_default_file() )!=NULL )
	{
		if( xmms_cfg_read_string( cfgfile, "xmmsd", "bind_to", &bt ) )
		{
			if( bind_to ) g_free( bind_to );
			bind_to=bt;
		}
		xmms_cfg_read_int( cfgfile, "xmmsd", "port", &port );
		xmms_cfg_read_boolean( cfgfile, "xmmsd", "require_login", &xc->require_login );
		if( xmms_cfg_read_string( cfgfile, "xmmsd", "user", &bt ) )
		{
			if( xc->user ) g_free( xc->user );
			xc->user=bt;
		}
		if( xmms_cfg_read_string( cfgfile, "xmmsd", "password", &bt ) )
		{
			if( xc->password ) g_free( xc->password );
			xc->password=bt;
		}
		xmms_cfg_free(cfgfile);
	}	
	if( port>65535 || port<1 ) port=8335;
	if( !bind_to ) bind_to=g_strdup( "127.0.0.1" );
}

void write_config( void )
{
	ConfigFile *cfgfile;

	if( ( cfgfile=xmms_cfg_open_default_file() )!=NULL )
	{
		xmms_cfg_write_string( cfgfile, "xmmsd", "bind_to", bind_to );
		xmms_cfg_write_int( cfgfile, "xmmsd", "port", port );
		xmms_cfg_write_boolean( cfgfile, "xmmsd", "require_login", xc->require_login );
		xmms_cfg_write_string( cfgfile, "xmmsd", "user", xc->user );
		xmms_cfg_write_string( cfgfile, "xmmsd", "password", xc->password );
		xmms_cfg_write_default_file( cfgfile );
		xmms_cfg_free(cfgfile);
	}
	else g_warning( "Couldn't open configfile\n" );
}

gboolean delete_event( GtkWidget *widget, GdkEvent *event, gpointer data )
{
	return FALSE;
}

void destroy (GtkWidget *widget, gpointer data)
{
	if( bind_to ) g_free( bind_to );
	bind_to=NULL;
	configure_win=NULL;
	xc_free();
}

void apply_clicked( GtkButton *button, gpointer user_data )
{
	gboolean restart=FALSE;
	gchar *gport=gtk_editable_get_chars( GTK_EDITABLE( port_entry ), 0, -1 );
	gchar *gbind_to=gtk_editable_get_chars( GTK_EDITABLE( bind_to_entry ), 0, -1 );
	gint lport=atoi( gport );
	if( lport>65535 || lport<1 ) lport=8335;
	if( strcmp( bind_to, gbind_to ) || lport!=port ) restart=TRUE;
	if( bind_to ) g_free( bind_to );
	bind_to=gbind_to;
	port=lport;

	xc->require_login=gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( require_login_checkbox ) );
	if( xc->user ) g_free( xc->user );
	xc->user=gtk_editable_get_chars( GTK_EDITABLE( user_entry ), 0, -1 );
	if( xc->password ) g_free( xc->password );
	xc->password=gtk_editable_get_chars( GTK_EDITABLE( password_entry ), 0, -1 );
	
	write_config();
	g_free( gport );
	if( restart ) restart_server();
}

void cancel_clicked( GtkButton *button, gpointer user_data )
{
	gtk_widget_destroy( configure_win );
}

void ok_clicked( GtkButton *button, gpointer user_data )
{
	apply_clicked( button, user_data );
	cancel_clicked( button, user_data );
}

void about_ok_clicked( GtkButton *button, gpointer user_data )
{
	gtk_widget_destroy( about_win );
}

void require_login_toggled( GtkToggleButton *togglebutton, gpointer user_data )
{
	gboolean rl=gtk_toggle_button_get_active( togglebutton );
	if( user_entry ) gtk_widget_set_sensitive( GTK_WIDGET( user_entry ), rl );
	if( user_label ) gtk_widget_set_sensitive( GTK_WIDGET( user_label ), rl );
	if( password_entry ) gtk_widget_set_sensitive( GTK_WIDGET( password_entry ), rl );
	if( password_label ) gtk_widget_set_sensitive( GTK_WIDGET( password_label ), rl );
}
	
void xc_free( void )
{
	if( !serv && !configure_win && xc )
	{
		if( xc->user ) g_free( xc->user );
		xc->user=NULL;
		if( xc->password ) g_free( xc->password );
		xc->password=NULL;
		g_free( xc );
		xc=NULL;
	}
}
