/*
    XMMSd - A web control plug-in for XMMS
    Copyright (C) 2002  Fredrik Rambris <boost@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* response.c - functions for creating the response that gets sent back to the
 * client
 */

#include <config.h>
#include "types.h"
#include "response.h"
#include "server.h"
#include "util.h"
#include <xmmsctrl.h>
#define FOOTER "<P><HR><A HREF=\"http://xmmsd.sourceforge.net/\" TARGET=\"_top\">XMMSd</A> " VERSION " by <A HREF=\"mailto:boost@users.sourceforge.net\">Fredrik Rambris</A>"

struct plnode
{
	gchar *playlist_file, *playlist_title;
	gint playlist_time;
};

struct xinfo
{
	gint xmms_session;
	gboolean is_playing, is_paused, is_main_win, is_pl_win, is_eq_win, is_running, is_repeat, is_shuffle;
	gint playlist_pos, playlist_length, output_time, left_vol, right_vol, main_vol, balance, playlist_time, rate, freq, nch;
	gchar *skin, *playlist_file, *playlist_title; /* Needs to be freed */
	gfloat preamp, *bands; /* Bands also neds to be freed */
	
};

/* Get much info in one go */
struct xinfo *xinfo_new( gint xmms_session )
{
	struct xinfo *xi=g_new0( struct xinfo, 1 );
	if( !xi ) return xi;
	xi->xmms_session=xmms_session;
	if( ( xi->is_running=xmms_remote_is_running( xi->xmms_session ) ) )
	{
		xi->is_playing=xmms_remote_is_playing( xi->xmms_session );
		xi->is_paused=xmms_remote_is_paused( xi->xmms_session );
		xi->is_main_win=xmms_remote_is_main_win( xi->xmms_session );
		xi->is_pl_win=xmms_remote_is_pl_win( xi->xmms_session );
		xi->is_eq_win=xmms_remote_is_eq_win( xi->xmms_session );
		xi->is_repeat=xmms_remote_is_repeat( xi->xmms_session );
		xi->is_shuffle=xmms_remote_is_shuffle( xi->xmms_session );
		xi->playlist_pos=xmms_remote_get_playlist_pos( xi->xmms_session );
		xi->playlist_length=xmms_remote_get_playlist_length( xi->xmms_session );
		xi->output_time=xmms_remote_get_output_time( xi->xmms_session );
		xmms_remote_get_volume( xi->xmms_session, &(xi->left_vol), &(xi->right_vol) );
		xi->main_vol=xmms_remote_get_main_volume( xi->xmms_session );
		xi->balance=xmms_remote_get_balance( xi->xmms_session );
		if( xi->playlist_pos>=0 )
		{
			xi->playlist_time=xmms_remote_get_playlist_time( xi->xmms_session, xi->playlist_pos );
			xi->playlist_file=xmms_remote_get_playlist_file( xi->xmms_session, xi->playlist_pos );
			xi->playlist_title=xmms_remote_get_playlist_title( xi->xmms_session, xi->playlist_pos );
		}
		xmms_remote_get_info( xi->xmms_session, &(xi->rate), &(xi->freq), &(xi->nch) );
		xi->skin=xmms_remote_get_skin( xi->xmms_session );
		xmms_remote_get_eq( xi->xmms_session, &(xi->preamp), &(xi->bands) );
	}
	else
	{
		g_free( xi );
		return NULL;
	}
	return xi;
}

void xinfo_free( struct xinfo *xi )
{
	if( !xi ) return;
	if( xi->skin ) g_free( xi->skin );
	xi->skin=NULL;
	if( xi->playlist_file ) g_free( xi->playlist_file );
	xi->playlist_file=NULL;
	if( xi->playlist_title ) g_free( xi->playlist_title );
	xi->playlist_title=NULL;
	if( xi->bands ) g_free( xi->bands );
	xi->bands=NULL;
	g_free( xi );
}

void xplaylist_freenode( gpointer data, gpointer user_data )
{
	struct plnode *pln=(struct plnode *)data;
	if( pln )
	{
		if( pln->playlist_file ) g_free( pln->playlist_file );
		if( pln->playlist_title ) g_free( pln->playlist_title );
		g_free( pln );
	}
}

void xplaylist_free( GList *list )
{
	if( !list ) return;
	g_list_foreach( list, xplaylist_freenode, NULL );
	g_list_free( list );
}

/* It takes about 0.6 seconds on my machine to get 2580 entries. */
GList *xplaylist_get( gint xmms_session )
{
	GList *list=NULL;
	struct plnode *pln;
	gint plength=xmms_remote_get_playlist_length( xmms_session );
	if( !plength ) return list;
	for( plength-- ; plength>=0 ; plength-- )
	{
		pln=g_new0( struct plnode, 1 );
		pln->playlist_file=xmms_remote_get_playlist_file( xmms_session, plength );
		pln->playlist_title=xmms_remote_get_playlist_title( xmms_session, plength );
		pln->playlist_time=xmms_remote_get_playlist_time( xmms_session, plength );
		list=g_list_prepend( list, pln );
	}
	return list;
}



/* A little take care of business (TM) */
gboolean tcb( struct Request *request, struct xinfo *xi )
{
	int i;
	gchar *action, *set, *to;
	action=(gchar *)g_hash_table_lookup( request->get_entries, "action" );
	set=(gchar *)g_hash_table_lookup( request->get_entries, "set" );
	to=(gchar *)g_hash_table_lookup( request->get_entries, "to" );

	if( !request || !xi ) return FALSE;

	if( action ) action=((GString *)action)->str;
	if( set ) set=((GString *)set)->str;
	if( to ) to=((GString *)to)->str;
	
	if( action )
	{
		if( !strcasecmp( action, "prev" ) ) xmms_remote_playlist_prev( xi->xmms_session );
		else if( !strcasecmp( action, "play" ) ) xmms_remote_play( xi->xmms_session );
		else if( !strcasecmp( action, "pause" ) ) xmms_remote_pause( xi->xmms_session );
#ifdef HAVE_XMMS_REMOTE_PLAY_PAUSE
		else if( !strcasecmp( action, "play_pause" ) ) xmms_remote_play_pause( xi->xmms_session );
#endif
		else if( !strcasecmp( action, "stop" ) ) xmms_remote_stop( xi->xmms_session );
		else if( !strcasecmp( action, "next" ) ) xmms_remote_playlist_next( xi->xmms_session );
		else return FALSE;

		return TRUE;
	}
	else if( set && to )
	{
		if( !strcasecmp( set, "volume" ) )
		{
			gint vol=atoi( to );
			xmms_remote_set_main_volume( xi->xmms_session, vol );
		}
		else if( !strcasecmp( set, "balance" ) )
		{
			gint balance=atoi( to );
			xmms_remote_set_balance( xi->xmms_session, balance );
		}
		else if( !strcasecmp( set, "ptime" ) )
		{
			gint jumptime=(atoi( to )*xi->playlist_time)/100;
			xmms_remote_jump_to_time( xi->xmms_session, jumptime );
		}
		else if( !strcasecmp( set, "time" ) )
		{
			gint jumptime=atoi( to );
			xmms_remote_jump_to_time( xi->xmms_session, jumptime );
		}
		else if( !strcasecmp( set, "shuffle" ) )
		{
			gboolean shuffle=(atoi( to )!=0);
			if( shuffle!=xi->is_shuffle ) xmms_remote_toggle_shuffle( xi->xmms_session );
		}
		else if( !strcasecmp( set, "repeat" ) )
		{
			gboolean repeat=(atoi( to )!=0);
			if( repeat!=xi->is_repeat ) xmms_remote_toggle_repeat( xi->xmms_session );
		}
		else if( !strcasecmp( set, "pos" ) )
		{
			gint pos=atoi( to );
			xmms_remote_set_playlist_pos( xi->xmms_session, pos );
		}
		else if( !strcasecmp( set, "eq_preamp" ) )
		{
			gfloat preamp=atof( to );
			xmms_remote_set_eq_preamp( xi->xmms_session, preamp );
		}
		else if( !strncasecmp( set, "eq_band[", 8 ) )
		{
			gint band;
			gfloat value=atof( to );
			sscanf( set, "eq_band[%d]", &band );
			xmms_remote_set_eq_band( xi->xmms_session, band, value );
		}
		else if( !strcasecmp( set, "main_win" ) )
		{
			gboolean show=atoi( to )!=0;
			if( xi->is_main_win!=show ) xmms_remote_main_win_toggle( xi->xmms_session, show );
		}
		else if( !strcasecmp( set, "eq_win" ) )
		{
			gboolean show=atoi( to )!=0;
			if( xi->is_eq_win!=show ) xmms_remote_eq_win_toggle( xi->xmms_session, show );
		}
		else if( !strcasecmp( set, "pl_win" ) )
		{
			gboolean show=atoi( to )!=0;
			if( xi->is_pl_win!=show ) xmms_remote_pl_win_toggle( xi->xmms_session, show );
		}
		else return FALSE;

		return TRUE;
	}
	return FALSE;
}

void add_titles( gint session, gchar *titles )
{
	GList *list=NULL;
	gchar *sptr=titles, *eptr=titles;
	gint num;
	if( !titles ) return;
	while( eptr[0] )
	{
		if( eptr[0]=='\n' || eptr[0]=='\r' )
		{
			eptr[0]='\0';
			if( sptr[0] ) /* We wont add empty strings */
			{
				stripws( sptr ); /* In case of spaces and such */
				list=g_list_append( list, sptr );
			}
			sptr=eptr+1;
		}
		eptr++;
	}
	if( eptr!=sptr && sptr[0] ) /* If we don't end with a newline the last string is catched here */
	{
		list=g_list_append( list, sptr );
	}
	xmms_remote_playlist_add( session, list );
	g_list_free( list );
}

void handle_http_response( struct Request *request )
{
	gint xmms_session=GetTagData( XMMSD_SESSION, 0, request->server->taglist );
	struct xinfo *xi=xinfo_new( xmms_session );
	int m, s, i, op;
	GList *plist;
	GList *node;
	struct plnode *pln;
	gint pos=0;
	gchar *get, *spos, *tstr, *referer;
	gboolean all, tcbd;
	gchar bands[]="eq_band[ ]", *what, *titles, *action;
	referer=(gchar *)g_hash_table_lookup( request->headers, "Referer" );
	if( !strcmp( request->path, "/" ) || !strcmp( request->path, "/index.html" ) )
	{
		g_string_sprintfa( request->response,
			"<HTML>\n"
			" <HEAD>\n"
			"  <TITLE>XMMSd</TITLE>\n"
			" </HEAD>\n"
			" <FRAMESET COLS=\"340,*\">\n"
			"  <FRAMESET ROWS=\"200,*\">\n"
			"   <FRAME SRC=\"main.html\" NAME=\"main\">\n"
			"   <FRAME SRC=\"eq.html\" NAME=\"eq\">\n"
			"  </FRAMESET>\n"
			"  <FRAME SRC=\"about.html\" NAME=\"playlist\">\n"
			" </FRAMESET>\n"
			" <NOFRAMES>\n"
			"  This site uses frames. There is <A HREF=\"/text.html\">another</A> interface for you\n"
			" </NOFRAMES>\n"
			"</HTML>"
		);
	}
	else if( !strcmp( request->path, "/main.html" )  )
	{
		if( xi )
		{
			if( tcb( request, xi ) )
			{
				usleep( 250000 );
				request_relocate( request, request->path );
			}
			else
			{
			m=xi->output_time/1000;
			s=m%60;
			m=(m-s)/60;
			op=100*xi->output_time/xi->playlist_time;
			g_string_sprintfa( request->response,
				"<!-- It's messy, I know!... That's why we have a /backend too -->\n"
				"<HTML>\n"
				"<HEAD>\n"
				"<LINK REL=\"stylesheet\" TYPE=\"text/css\" HREF=\"/style.css\">\n"
				"<META http-equiv=\"refresh\" content=\"%d,/main.html\">\n"
				"<TITLE>XMMSd</TITLE>\n"
				"</HEAD>\n"
				"<BODY>\n"
				"<TABLE CELLSPACING=\"4\" CELLPADDING=\"2\" STYLE=\"border: outset 2px #404040; background: #404040;\">\n"
				"<TR><TD CLASS=\"time\" ROWSPAN=\"2\" ALIGN=\"center\">%02d:%02d</TD><TD STYLE=\"border: inset 1px #404040; background: #000000; color: #2498ff;\"><TABLE WIDTH=\"100%%\"><TR><TD COLSPAN=\"3\" STYLE=\"color: #2498ff; font-size: 8pt\">%s</TD></TR><TR><TD STYLE=\"color: #2498ff; font-size: 8pt\">%d kbps</TD><TD STYLE=\"color: #2498ff; font-size: 8pt\">%d kHz</TD><TD ALIGN=\"right\"><SPAN CLASS=\"inv\">%s</SPAN></TD></TR></TABLE></TD></TR>\n"
				"<TR><TD><TABLE WIDTH=\"100%%\"><TR><TD CLASS=\"slider\">", ((xi->playlist_time-xi->output_time)/1000)+2, m, s, xi->playlist_title, (int)(xi->rate/1000), (int)(xi->freq/1000), (xi->nch==2?"STEREO":(xi->nch==1?"MONO":""))
			);
			for( i=0; i<102 ; i+=5 )
			{
				if( i>=xi->main_vol && i<(xi->main_vol+5) ) g_string_sprintfa( request->response, "<SPAN CLASS=\"knob\">&nbsp;</SPAN>" );
				else g_string_sprintfa( request->response, "<A HREF=\"%s?set=volume&to=%d\">&nbsp;</A>", request->path, i );
			}

			g_string_sprintfa( request->response, "</TD><TD CLASS=\"slider\">" );

			for( i=-100; i<102 ; i+=10 )
			{

				if( i>=xi->balance && i<(xi->balance+10) ) g_string_sprintfa( request->response, "<SPAN CLASS=\"knob\">&nbsp;</SPAN>" );
				else g_string_sprintfa( request->response, "<A HREF=\"%s?set=balance&to=%d\">&nbsp;</A>", request->path, i );
			}

			g_string_sprintfa( request->response, "</TD><TD ALIGN=\"right\"><A HREF=\"eq.html\" TARGET=\"eq\" CLASS=\"tbut\">EQ</A> <A HREF=\"/playlist.html\" TARGET=\"playlist\" CLASS=\"tbut\">PL</A></TD></TR></TABLE></TD></TR><TR><TD COLSPAN=\"2\" CLASS=\"slider\">\n"
			);
			for( i=0; i<100 ; i++ )
			{
				if( i==op ) g_string_sprintfa( request->response, "<SPAN CLASS=\"knob\"><A HREF=\"/?set=ptime&to=%d\">&nbsp;</A></SPAN>" );
				else g_string_sprintfa( request->response, "<A HREF=\"%s?set=ptime&to=%d\">&nbsp;</A>", request->path, i );
			}
			g_string_sprintfa( request->response,
				"</TD></TR>\n"
				"<TR><TD COLSPAN=\"2\"><TABLE WIDTH=\"100%%\"><TR><TD><A CLASS=\"but\" HREF=\"%s?action=prev\">|&lt;</A> <A CLASS=\"but\" HREF=\"%s?action=play\">&gt;</A> <A CLASS=\"but\" HREF=\"%s?action="
#ifdef HAVE_XMMS_REMOTE_PLAY_PAUSE
				"play_pause"
#else
				"pause"
#endif
				"\">||</A> <A CLASS=\"but\" HREF=\"%s?action=stop\">Stop</A> <A CLASS=\"but\" HREF=\"%s?action=next\">&gt;|</A></TD><TD ALIGN=\"right\"><A CLASS=\"%stbut\" HREF=\"%s?set=shuffle&to=%d\">Rand</A> <A CLASS=\"%stbut\" HREF=\"%s?set=repeat&to=%d\">Rep</A> <A HREF=\"/about.html\" TARGET=\"playlist\" CLASS=\"but\">X</A></TD></TR></TABLE></TD></TR>\n"
				"</TABLE>\n</BODY>\n</HTML>",
				request->path,
				request->path,
				request->path,
				request->path,
				request->path,
				(xi->is_shuffle?"s":""),
				request->path,
				(xi->is_shuffle?0:1),
				(xi->is_repeat?"s":""),
				request->path,
				(xi->is_repeat?0:1)
			);
			}
		}
		else
		{
			g_string_sprintfa( request->response,
				"<HTML>\n"
				"<HEAD>\n"
				"<LINK REL=\"stylesheet\" TYPE=\"text/css\" HREF=\"/style.css\">\n"
				"<TITLE>XMMSd</TITLE>\n"
				"</HEAD>\n"
				"<BODY>\n"
				"XMMS is not running\n"
			);
		}
	}
	else if( !strcmp( request->path, "/style.css" ) )
	{
		g_string_sprintf( request->content_type, "text/css" );
		g_string_sprintfa( request->response,
			"BODY, TD { font-family: Verdana,sans-serif; font-size: 11pt; }\n"
			"BODY { background: #52568b; color: #ffffff; }\n"
			"A:link,A:visited { color: #c0c0c0; text-decoration: none; font-weight: bold }\n"
			"A:link:hover,A:visited:hover { color: #c0c0ff; text-decoration: underline }\n"
			"A.but { background: #505050; color: #2498ff; border: outset 2px #505050; font-weight: bold; font-size: 9pt; padding: 1px 4px 1px 4px }\n"
			"A.but:hover { text-decoration: none; border: outset 2px #606060; background: #606060; }\n"
			"A.but:active { border: inset 2px #505050; background: #505050; }\n"
			".time { border: inset 1px #404040; background: #000000; font-family: monospace; font-size: 16pt; font-weight: bold; color: #2498ff; }\n"
			"A.tbut { background: #505050; color: #ffffff; border: outset 2px #505050; font-weight: normal; font-size: 8pt; padding: 1px 4px 1px 4px }\n"
			"A.tbut:hover { text-decoration: none; border: outset 2px #606060; background: #606060; }\n"
			"A.tbut:active { border: inset 2px #505050; background: #505050; }\n"
			"A.stbut { margin: 1px; background: #505050; color: #ffffff; border: inset 1px #505050; font-weight: normal; font-size: 8pt; padding: 2px 5px 2px 5px }\n"
			"A.stbut:hover { text-decoration: none; border: inset 1px #606060; background: #606060; }\n"
			"A.stbut:active { border: inset 2px #505050; background: #505050; padding: 1px 4px 1px 4px; }\n"
			"TD.slider { text-align: center }\n"
			"TD.slider A { font-size: 6pt; text-decoration: line-through }\n"
			"TD.slider A:hover { text-decoration: none; }\n"
			"SPAN.knob { border: outset 1px #2498ff; background: #2498ff; }\n"
			"SPAN.inv { font-size: 7pt; font-weight: bold; background: #2498ff; color: #000000; padding: 0px 2px 0px 2px }\n"
			"DIV.playlist TD { font-size: 9pt; background: #000000 }\n"
			"DIV.playlist TD A { font-weight: normal }\n"
		);
	}
	else if( !strcmp( request->path, "/about.html" ) )
	{
		g_string_sprintfa( request->response,
			"<HTML>\n"
			"<HEAD>\n"
			"<LINK REL=\"stylesheet\" TYPE=\"text/css\" HREF=\"/style.css\">\n"
			"<TITLE>XMMSd</TITLE>\n"
			"</HEAD>\n"
			"<BODY>\n"
			"<H1>XMMS</H1>\n"
			"<A HREF=\"http://xmmsd.sourceforge.net\" TARGET=\"_top\">XMMSd</A> (X Multi Media System Daemon) is a web interface for controling\n"
			"<A HREF=\"http://www.xmms.com/\" TARGET=\"_top\">XMMS</A> developed by\n"
			"<A HREF=\"http://fredrik.rambris.com/\" TARGET=\"_top\">Fredrik Rambris</A>. It is\n"
			"built pimary as a plug-in for XMMS but can also be started from command line.\n"
			"As well as offering a GUI-like interface (HTML4+CSS) a backend interface which\n"
			"try to <I>webify</I> XMMS' own interface and returning compact results makes\n"
			"it suitable both for end users and other front ends.<P>\n"
			"The approach with HTTP makes it accessible by other users and other computers\n"
			"(if you want). This makes it more easy to use it as a back-end from languages\n"
			"such as PHP and Java (although there are unix domain sockets available for\n"
			"Java). HTTP is (rather) well documented, offers good portability and ofcause\n"
			"is very wide spread.\n"
			FOOTER
			"</BODY>\n"
			"</HTML>"
		);
	}
	else if( !strcmp( request->path, "/playlist.html" ) )
	{
		g_string_sprintfa( request->response,
			"<HTML>\n"
			"<HEAD>\n"
			"<LINK REL=\"stylesheet\" TYPE=\"text/css\" HREF=\"/style.css\">\n"
			"<TITLE>XMMSd - playlist</TITLE>\n"
			"</HEAD>\n"
			"<BODY>\n"
		);
		if( ( plist=xplaylist_get( xmms_session ) ) )
		{
			g_string_sprintfa( request->response, "<DIV CLASS=\"playlist\"><TABLE WIDTH=\"100%%\" CELLSPACING=\"0\">\n" );
			for( node=plist; node; node=node->next )
			{
				pln=(struct plnode *)node->data;
				if( pln )
				{
					m=pln->playlist_time/1000;
					s=m%60;
					m=(m-s)/60;
					g_string_sprintfa( request->response, "<TR><TD VALIGN=\"top\" ALIGN=\"right\">%d.</TD><TD><A HREF=\"/main.html?set=pos&to=%d\" TARGET=\"main\">%s</A></TD><TD VALIGN=\"top\" ALIGN=\"right\">%d:%02d</TD></TR>\n", pos, pos, pln->playlist_title, m, s );
				}
				pos++;
			}
			g_string_sprintfa( request->response, "</TABLE></DIV>\n" );
			xplaylist_free( plist );
		}
		else
		{
			g_string_sprintfa( request->response, "ERROR" );
		}
		g_string_sprintfa( request->response,
			"</BODY>\n"
			"</HTML>"
		);
	}
	else if( !strcmp( request->path, "/eq.html" ) )
	{
		g_string_sprintfa( request->response,
			"<HTML>\n"
			"<HEAD>\n"
			"<LINK REL=\"stylesheet\" TYPE=\"text/css\" HREF=\"/style.css\">\n"
			"<TITLE>XMMSd</TITLE>\n"
			"</HEAD>\n"
			"<BODY>\n"
			"<TABLE CELLSPACING=\"4\" CELLPADDING=\"2\" STYLE=\"border: outset 2px #404040; background: #404040;\">\n"
			"<TR><TD>In time this will be an equalizer control</TD></TR>\n"
			"</TABLE>\n"
			"</BODY>\n"
			"</HTML>"
		);
	}
	else if( !strcmp( request->path, "/text.html" ) )
	{
		if( xi )
		{
			if( tcb( request, xi ) )
			{
				/* Instead of makeing a meta refresh we sleep here instead and relocate */
				usleep( 250000 );
				request_relocate( request, request->path );
			}
			else
			{
				g_string_sprintfa( request->response,
					"<HTML>\n"
					"<HEAD>\n"
					"<TITLE>XMMSd</TITLE>\n"
					"</HEAD>\n"
					"<BODY>\n"
					"Currently playing: <B>%s</B><P>\n"
					"[ <A HREF=\"/text.html?action=prev\">Prev</A> | <A HREF=\"/text.html?action=play\">Play</A> | <A HREF=\"/text.html?action="
#ifdef HAVE_XMMS_REMOTE_PLAY_PAUSE
					"play_pause"
#else
					"pause"
#endif
					"\">%s</A> | <A HREF=\"/text.html?action=stop\">Stop</A> | <A HREF=\"/text.html?action=next\">Next</A> ]<BR>\n"
					"</BODY>\n"
					"</HTML>",
					xi->playlist_title,
					(xi->is_paused?"Unpause":"Pause")
				);
			}
		}
		else
		{
			g_string_sprintfa( request->response,
				"<HTML>\n"
				"<HEAD>\n"
				"<TITLE>XMMSd</TITLE>\n"
				"</HEAD>\n"
				"<BODY>\n"
				"XMMS isn't running\n"
				"</BODY>\n"
				"</HTML>"
			);
		}
	}
	else if( !strcmp( request->path, "/backend" ) )
	{
		request_relocate( request, "/backend/" );
	}
	else if( !strcmp( request->path, "/backend/" ) )
	{
		get=(gchar *)g_hash_table_lookup( request->get_entries, "get" );
		if( get ) get=((GString *)get)->str;

		if( tcb( request, xi ) )
		{
			g_string_sprintfa( request->response, "<HTML><HEAD><META http-equiv=\"refresh\" content=\"1,%s\"</HEAD><BODY>\nCommand executed. It's now dead.\n</BODY></HTML>", (referer?referer:request->path) );
		}
		else if( get )
		{
			tstr=NULL;
			spos=(gchar *)g_hash_table_lookup( request->get_entries, "pos" );
			if( spos ) spos=((GString *)spos)->str;
			if( spos ) pos=atoi( spos );
			else if( xi ) pos=xi->playlist_pos;
			else pos=-1;
			g_string_sprintf( request->content_type, "text/plain; charset=iso-8859-1" );
			if( !strcasecmp( get, "all" ) ) all=TRUE;
			else all=FALSE;
			if( xi )
			{
				if( all || !strcasecmp( get, "is_running" ) ) g_string_sprintfa( request->response, "is_running=1\n" );
				if( all || !strcasecmp( get, "is_playing" ) ) g_string_sprintfa( request->response, "is_playing=%d\n", xi->is_playing );
				if( all || !strcasecmp( get, "is_paused" ) ) g_string_sprintfa( request->response, "is_paused=%d\n", xi->is_paused );
				if( all || !strcasecmp( get, "is_main_win" ) ) g_string_sprintfa( request->response, "is_main_win=%d\n", xi->is_main_win );
				if( all || !strcasecmp( get, "is_pl_win" ) ) g_string_sprintfa( request->response, "is_pl_win=%d\n", xi->is_pl_win );
				if( all || !strcasecmp( get, "is_eq_win" ) ) g_string_sprintfa( request->response, "is_eq_win=%d\n", xi->is_eq_win );
				if( all || !strcasecmp( get, "is_repeat" ) ) g_string_sprintfa( request->response, "is_repeat=%d\n", xi->is_repeat );
				if( all || !strcasecmp( get, "is_shuffle" ) ) g_string_sprintfa( request->response, "is_shuffle=%d\n", xi->is_shuffle );
				if( all || !strcasecmp( get, "playlist_pos" ) ) g_string_sprintfa( request->response, "playlist_pos=%d\n", xi->playlist_pos );
				if( all || !strcasecmp( get, "playlist_length" ) ) g_string_sprintfa( request->response, "playlist_length=%d\n", xi->playlist_length );
				if( all || !strcasecmp( get, "playlist_time" ) ) g_string_sprintfa( request->response, "playlist_time=%d\n", xmms_remote_get_playlist_time( xi->xmms_session, pos ) );
				if( all || !strcasecmp( get, "output_time" ) ) g_string_sprintfa( request->response, "output_time=%d\n", xi->output_time );
				if( all || !strcasecmp( get, "volume" ) ) g_string_sprintfa( request->response, "volume=%d\n", xi->main_vol );
				if( all || !strcasecmp( get, "balance" ) ) g_string_sprintfa( request->response, "balance=%d\n", xi->balance );
				if( all || !strcasecmp( get, "rate" ) ) g_string_sprintfa( request->response, "rate=%d\n", xi->rate );
				if( all || !strcasecmp( get, "freq" ) ) g_string_sprintfa( request->response, "freq=%d\n", xi->freq );
				if( all || !strcasecmp( get, "nch" ) ) g_string_sprintfa( request->response, "nch=%d\n", xi->nch );
				if( all || !strcasecmp( get, "skin" ) ) g_string_sprintfa( request->response, "skin=%s\n", (xi->skin?xi->skin:"") );
				if( all || !strcasecmp( get, "playlist_file" ) ) g_string_sprintfa( request->response, "playlist_file=%s\n", (tstr=xmms_remote_get_playlist_file( xi->xmms_session, pos ) ) );
				if( tstr ) g_free( tstr );
				tstr=NULL;
				if( all || !strcasecmp( get, "playlist_title" ) ) g_string_sprintfa( request->response, "playlist_title=%s\n", (tstr=xmms_remote_get_playlist_title( xi->xmms_session, pos ) ) );
				if( tstr ) g_free( tstr );
				tstr=NULL;
				if( all || !strcasecmp( get, "eq_preamp" ) ) g_string_sprintfa( request->response, "eq_preamp=%.2f\n", xi->preamp );
				for( i=0; i<10 ; i++ )
				{
					sprintf( bands, "eq_band[%d]", i );
					if( all || !strcasecmp( get, bands ) ) g_string_sprintfa( request->response, "%s=%.2f\n", bands, xi->bands[i] );
				}
				if( !strcasecmp( get, "playlist" ) ) /* This must be specificly asked for */
				{
					if( ( plist=xplaylist_get( xmms_session ) ) )
					{
						for( node=plist; node; node=node->next )
						{
							pln=(struct plnode *)node->data;
							if( pln )
							{
								g_string_sprintfa( request->response, "%s\t%s\t%d\n", pln->playlist_file, pln->playlist_title, pln->playlist_time );
							}
						}
						xplaylist_free( plist );
					}
					else
					{
						g_string_sprintfa( request->response, "ERROR" );
					}
				}
			}
			else
			{
				g_string_sprintfa( request->response, "ERR=XMMS isn't running\n" );
			}
		}
		else
		{
			g_string_sprintfa( request->response,
				"<HTML><HEAD><TITLE>XMMSd</TITLE></HEAD><BODY>\n"
				"This is <I>backend</I>. Here are no fancy stuff. See <A HREF=\"help.html\">help</A> for info on how to use this interface.<P>XMMS %s running.\n"
				FOOTER
				"</BODY></HTML>", (xi?"is":"isn't")
			);
		}
	}
	else if( !strcmp( request->path, "/backend/playlist.html" ) )
	{
		if( !xi )
		{
			g_string_sprintfa( request->response, "ERR=XMMS isn't running\n" );
		}
		else
		{
			tcbd=FALSE;
			if( request->method==METHOD_POST && request->postdata )
			{
				action=(gchar *)g_hash_table_lookup( request->post_entries, "action" );
				titles=(gchar *)g_hash_table_lookup( request->post_entries, "titles" );
				spos=(gchar *)g_hash_table_lookup( request->post_entries, "pos" );
				if( titles ) titles=((GString *)titles)->str;
				if( action ) action=((GString *)action)->str;
				if( spos ) spos=((GString *)spos)->str;

				if( action )
				{
					tcbd=TRUE;
					if( !strcasecmp( action, "enqueue" ) && titles ) add_titles( xi->xmms_session, titles );
					else if( !strcasecmp( action, "replace" ) && titles )
					{
						xmms_remote_playlist_clear( xi->xmms_session );
						add_titles( xi->xmms_session, titles );
					}
					else if( !strcasecmp( action, "delete" ) && spos )
					{
						pos=atoi( spos );
						xmms_remote_playlist_delete( xi->xmms_session, pos );
					}
					else if( !strcasecmp( action, "clear" ) )
					{
						xmms_remote_playlist_clear( xi->xmms_session );
					}
					else
					{
						tcbd=FALSE;
					}
				}

			}
			else
			{
				action=(gchar *)g_hash_table_lookup( request->get_entries, "action" );
				if( action ) action=((GString *)action)->str;
				spos=(gchar *)g_hash_table_lookup( request->post_entries, "pos" );
				if( spos ) spos=((GString *)spos)->str;

				if( action )
				{
					tcbd=TRUE;
					if( !strcasecmp( action, "clear" ) )
					{
						xmms_remote_playlist_clear( xi->xmms_session );
					}
					else if( !strcasecmp( action, "delete" ) && spos )
					{
						pos=atoi( spos );
						xmms_remote_playlist_delete( xi->xmms_session, pos );
					}
					else
					{
						tcbd=FALSE;
					}
				}

			}
			if( tcbd )
			{
				g_string_sprintfa( request->response, "<HTML><HEAD><META http-equiv=\"refresh\" content=\"1,%s\"</HEAD><BODY>\nCommand executed. It's now dead.\n</BODY></HTML>", (referer?referer:request->path) );
			}
			else
			{
				g_string_sprintfa( request->response,
				"<HTML>\n"
				"<HEAD>\n"
				"<TITLE>XMMSd</TITLE>\n"
				"</HEAD>\n"
				"<BODY>\n"
				"This interface is better reachable through POST method with url encoded form data apart from the action <B>clear</B> and <B>delete</B>.<P>\n"
				"See <A HREF=\"/backend/help.html#playlist\">help</A> for more info.<P>\n"
				"You can sneak-use it here though by this form although it is not very usable<BR>\n"
				"<FORM METHOD=\"post\" ACTION=\"/backend/playlist.html\">\n"
				"<SELECT NAME=\"action\">\n"
				"<OPTION VALUE=\"enqueue\">Enqueue</OPTION>\n"
				"<OPTION VALUE=\"replace\">Replace</OPTION>\n"
				"</SELECT><BR>\n"
				"<TEXTAREA COLS=\"50\" ROWS=\"10\" NAME=\"titles\"></TEXTAREA><BR>\n"
				"<INPUT TYPE=\"submit\" VALUE=\"Submit\">\n"
				"</FORM>\n"
				"</BODY>\n"
				"</HTML>"
			);

			}
		}
	}
	else if( !strcmp( request->path, "/backend/help.html" ) )
	{
		g_string_sprintfa( request->response,
			"<HTML>\n"
			"<HEAD>\n"
			"<LINK REL=\"stylesheet\" TYPE=\"text/css\" HREF=\"/style.css\">\n"
			"<TITLE>XMMSd - backend help</TITLE>\n"
			"</HEAD>\n"
			"<BODY>\n"
			"<H1>Backend help</H1>\n"
			"You use this by using three commands. The commands are <A HREF=\"#action\">action</A> <A HREF=\"#get\">get</A> <A HREF=\"#set\">set</A>.<P>\n"
			"<H3><A NAME=\"action\">ACTION</A></H3>\n"
			"You use action like this http://<I>hostname</I>:<I>port</I>/backend/?action=<I>action</I><BR>\n"
			"We just concentrate on the stuff after the question sign here. The following actions exists\n"
			"<DL>\n"
			"<DT><A HREF=\"/backend/?action=prev\">prev</A>\n"
			"<DD>skips back to the previous song\n"
			"<DT><A HREF=\"/backend/?action=next\">next</A>\n"
			"<DD>skips to the next song\n"
			"<DT><A HREF=\"/backend/?action=play\">play</A>\n"
			"<DD>starts/restart/resume playing\n"
			"<DT><A HREF=\"/backend/?action=stop\">stop</A>\n"
			"<DD>stops playing\n"
			"<DT><A HREF=\"/backend/?action=pause\">pause</A>\n"
			"<DD>pauses play\n"
			"<DT><A HREF=\"/backend/?action=play_pause\">play_pause</A>\n"
			"<DD>alternates between play and pause\n"
#ifndef HAVE_XMMS_REMOTE_PLAY_PAUSE
			"The version of XMMS installed doesn't support this function and therefor is not available\n"
#endif
			"</DL><P>\n"
			"<H3><A NAME=\"get\">GET</A></H3>\n"
			"Get is used to get information from XMMS and used like this:<BR>\n"
			"http://<I>hostname</I>:<I>port</I>/backend/?get=<I>what</I>&amp;pos=<I>pos</I><BR>\n"
			"You can leave out pos if you don't want to specify it.<BR>\n"
			"<I>what</I> can be one of the following:<BR>\n"
			"<DL>\n"
			"<DT><A HREF=\"/backend/?get=all\">all</A>\n"
			"<DD>returns all information except the playlist\n"
			"<DT><A HREF=\"/backend/?get=is_running\">is_running</A>\n"
			"<DD>1 or 0. Is XMMS running\n"
			"<DT><A HREF=\"/backend/?get=is_playing\">is_playing</A>\n"
			"<DD>1 or 0. Is something playing\n"
			"<DT><A HREF=\"/backend/?get=is_paused\">is_paused</A>\n"
			"<DD>1 or 0. Is playing paused\n"
			"<DT><A HREF=\"/backend/?get=is_main_win\">is_main_win</A>\n"
			"<DD>1 or 0. Is the main window open\n"
			"<DT><A HREF=\"/backend/?get=is_pl_win\">is_pl_win</A>\n"
			"<DD>Is playlist window open\n"
			"<DT><A HREF=\"/backend/?get=is_eq_win\">is_eq_win</A>\n"
			"<DD>1 or 0. Is the equalizer window open\n"
			"<DT><A HREF=\"/backend/?get=is_repeat\">is_repeat</A>\n"
			"<DD>1 or 0. Is repeat on or off\n"
			"<DT><A HREF=\"/backend/?get=is_shuffle\">is_shuffle</A>\n"
			"<DD>1 or 0. Is shuffle on or off\n"
			"<DT><A HREF=\"/backend/?get=playlist_pos\">playlist_pos</A>\n"
			"<DD>Position of the currently playing song in the playlist. Starting from 0.\n"
			"<DT><A HREF=\"/backend/?get=playlist_length\">playlist_length</A>\n"
			"<DD>Number of titles in the playlist\n"
			"<DT><A HREF=\"/backend/?get=playlist_time\">playlist_time</A>\n"
			"<DD>The length of the song on <I>pos</I> or the current. In ms.\n"
			"<DT><A HREF=\"/backend/?get=playlist_file\">playlist_file</A>\n"
			"<DD>The filename of the song on <I>pos</I> or the current\n"
			"<DT><A HREF=\"/backend/?get=playlist_title\">playlist_title</A>\n"
			"<DD>The title if the song on <I>pos</I> or the current\n"
			"<DT><A HREF=\"/backend/?get=output_time\">output_time</A>\n"
			"<DD>How far the current song has played. In ms\n"
			"<DT><A HREF=\"/backend/?get=volume\">volume</A>\n"
			"<DD>0 to 100. The current volume\n"
			"<DT><A HREF=\"/backend/?get=balance\">balance</A>\n"
			"<DD>-100 to 100. Balance between left and right channel\n"
			"<DT><A HREF=\"/backend/?get=rate\">rate</A>\n"
			"<DD>The bitrate on the current song. I.e. 192000 128000 etc.\n"
			"<DT><A HREF=\"/backend/?get=nch\">nch</A>\n"
			"<DD>Number of channels. 1 for mono, 2 for stereo or 0 for unknown\n"
			"<DT><A HREF=\"/backend/?get=skin\">skin</A>\n"
			"<DD>Filename of current skin or empty if it's the default skin\n"
			"<DT><A HREF=\"/backend/?get=eq_preamp\">eq_preamp</A>\n"
			"<DD>-20 to 20. Preamp value in the equalizer. In dB\n"
			"<DT><A HREF=\"/backend/?get=eq_band[0]\">eq_band[&lt;0-9&gt;]</A>\n"
			"<DD>-20 to 20 the amp value for each band\n"
			"<DT><A HREF=\"/backend/?get=playlist\">playlist</A>\n"
			"<DD>Return the whole playlist in a tab separated list: file title time\n"
			"</DL><P>\n"
			"<H3><A NAME=\"set\">SET</A></H3>\n"
			"Set modifies values in XMMS with the following syntax<BR>\n"
			"http://<I>hostname</I>:<I>port</I>/backend/?set=<I>what</I>&to=<I>value</I><BR>\n"
			"You can set the following values:<BR>\n"
			"<DL>\n"
			"<DT>volume\n"
			"<DD>The main volume (0-100)\n"
			"<DT>balance\n"
			"<DD>Balance (-100 - 100)\n"
			"<DT>time\n"
			"<DD>Jump to specified time in the current song. In ms\n"
			"<DT>ptime\n"
			"<DD>Jump to specified time in the current song. In percent\n"
			"<DT>shuffle\n"
			"<DD>Set shuffle to 1 or 0\n"
			"<DT>repeat\n"
			"<DD>Set repeat to 1 or 0\n"
			"<DT>pos\n"
			"<DD>Skip to the specified position in the playlist\n"
			"<DT>eq_preamp\n"
			"<DD>Set equalizer preamp\n"
			"<DT>eq_band[&lt;0-9&gt;]\n"
			"<DD>Set equalizer amplification per band\n"
			"<DT>main_win\n"
			"<DD>Show (1) or hide (0) the main window\n"
			"<DT>pl_win\n"
			"<DD>Show (1) or hide (0) the playlist window\n"
			"<DT>eq_win\n"
			"<DD>Show (1) or hide (0) the equalizer window\n"
			"</DL><P>\n"
			"<H3><A NAME=\"playlist\">PLAYLIST</A></H3>\n"
			"Under the url http://<I>hostname</I>:<I>port</I>/backend/playlist.html you can modify the playlist.<BR>\n"
			"Most is done with the method POST but you can clear the playlist with GET. <B>/backend/playlist.html?action=clear</B>.<P>\n"
			"When posting data you send <B>action</B>=<I>action</I>&amp;<B>titles</B>=<I>titles</I>&amp;<B>pos</B>=<I>pos</I><BR>\n"
			"Be sure to url encode it all before sending it<BR>\n"
			"<B>action</B> can be one of the following:<BR>\n"
			"<DL>\n"
			"<DT>clear\n"
			"<DD>Clears the playlist and stop playing.<BR>\n"
			"<DT>enqueue\n"
			"<DD>Adds the files or URLs supplied in <B>titles</B> (separated by newlines; \\n)\n"
			"<DT>replace\n"
			"<DD>Replace the current playlist with the files or URLs supplied in <B>titles</B> (separated by newlines; \\n)\n"
			"<DT>delete\n"
			"<DD>Removes the line <I>pos</I> from playlist\n"
			"</DL>\n"
			FOOTER
			"</BODY></HTML>"
		);
	}
	else
	{
		request->ret_code=404;
	}
	if( xi ) xinfo_free( xi );
}

gboolean authenticate( struct Request *req )
{
	g_string_sprintf( req->response, "<HTML><HEAD><TITLE>XMMSd</TITLE></HEAD><BODY>You are note authorized to access this site</BODY></HTML>" ); /* Newlines are for loosers ;-) */
	req->ret_code=401;
	req->response_headers=g_list_prepend( req->response_headers, g_strdup("WWW-Authenticate: Basic realm=\"XMMSd\"") );
	return FALSE;
}

gboolean auth_handler( struct Request *request )
{
	struct xmmsd_config *xc=request->server->user_data;
	if( !xc ) return( authenticate( request ) ); /* We shut out rather than let in on error */
	if( xc->require_login && xc->user && strlen( xc->user ) && xc->password && strlen( xc->password ) )
	{
		if( !request->auth_user || !request->auth_pass ) return( authenticate( request ) );
		if( strcmp( xc->user, request->auth_user ) || strcmp( xc->password, request->auth_pass ) ) return( authenticate( request ) );
	}
	else return TRUE; /* However... if we haven't set all options we assume that we grant everybody */
}
