/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef TAGLIST_H
#define TAGLIST_H

#include <stdio.h>
#include <glib.h>

/* This is a reverse engineered port of the TagList concept found in AmigaOS
 * release 2 and later. As I haven't invented it I don't hold any copyrights.
 * Port by Fredrik Rambris <fredrik@rambris.com>.
 * Copyright (c) Amiga, Inc. */

/* Update: I have glib-ified it somewhat. I no longer use ULONG for everything
 * but rather gint, guint and gboolean from glib. */

/*****************************************************************************/


/* Tags are a general mechanism of extensible data arrays for parameter
 * specification and property inquiry. In practice, tags are used in arrays,
 * or chain of arrays.
 *
 */

typedef guint Tag;

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

struct TagItem
{
    Tag  ti_Tag;	/* identifies the type of data */
    gint ti_Data;	/* type-specific data	       */
};

/* constants for Tag.ti_Tag, control tag values */
#define TAG_DONE   (0L)	  /* terminates array of TagItems. ti_Data unused */
#define TAG_END	   (0L)   /* synonym for TAG_DONE			  */
#define	TAG_IGNORE (1L)	  /* ignore this item, not end of array		  */
#define	TAG_MORE   (2L)	  /* ti_Data is pointer to another array of TagItems
			   * note that this tag terminates the current array
			   */
#define	TAG_SKIP   (3L)	  /* skip this and the next ti_Data items	  */

/* differentiates user tags from control tags */
#define TAG_USER   ((guint)(1L<<31))

/* If the TAG_USER bit is set in a tag number, it tells utility.library that
 * the tag is not a control tag (like TAG_DONE, TAG_IGNORE, TAG_MORE) and is
 * instead an application tag. "USER" means a client of utility.library in
 * general, including system code like Intuition or ASL, it has nothing to do
 * with user code.
 */


/*****************************************************************************/


/* Tag filter logic specifiers for use with FilterTagItems() */
#define TAGFILTER_AND 0		/* exclude everything but filter hits	*/
#define TAGFILTER_NOT 1		/* exclude only filter hits		*/


/*****************************************************************************/


/* Mapping types for use with MapTags() */
#define MAP_REMOVE_NOT_FOUND 0	/* remove tags that aren't in mapList */
#define MAP_KEEP_NOT_FOUND   1	/* keep tags that aren't in mapList   */


/*****************************************************************************/

struct TagItem *AllocateTagItems( gint numTags );
void FreeTagItems( struct TagItem *tagList );
struct TagItem *NextTagItem( struct TagItem **tagItemPtr );
void RefreshTagItemClones( struct TagItem *clone, struct TagItem *original );
struct TagItem *CloneTagItems( struct TagItem *original );
struct TagItem *FindTagItem( Tag tagValue, struct TagItem *tagList );
gint GetTagData( Tag tagValue, gint defaultVal, struct TagItem *tagList );
gboolean TagInArray( Tag tagValue, Tag *tagArray );
gint FilterTagItems( struct TagItem *tagList, Tag *filterArray, gboolean logic );
void ApplyTagChanges( struct TagItem *list, struct TagItem *changeList );
void FilterTagChanges( struct TagItem *changeList, struct TagItem *originalList, gboolean apply );
void MapTags( struct TagItem *tagList, struct TagItem *mapList, gint mapType );
guint PackBoolTags( guint initialFlags, struct TagItem *tagList, struct TagItem *boolMap );

/*** NOT IMPLEMENTED ***/
// gint PackStructureTags( void *pack, gint *packTable, struct TagItem *tagList );
// gint UnpackStructureTags( void *pack, gint *packTable, struct TagItem *tagList );
 
#endif
