#ifndef TYPES_H
/*
    XMMSd - A web control plug-in for XMMS
    Copyright (C) 2002  Fredrik Rambris <boost@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#define TYPES_H

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <glib.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "taglist.h"

struct Entry
{
	gchar key[128];
	gchar val[512];
};

#define METHOD_GET 1
#define METHOD_POST 2

/* Prototype to allow a pointer in response_handler */
struct Request;

struct Server
{
	pthread_t thread; /* Thread of server */
	gboolean running; /* Set to false to shut down the server politely */
	struct sockaddr_in myaddr; /* The whole address we're bound to */
	in_addr_t address; /* The address we're bound to. Just the actual address */
	gint port; /* The port we're listening to */
	gint servsock; /* Inbound server socket*/
	struct Request **clients; /* An array of clients being served */
	pthread_mutex_t clients_mutex; /* Get a lock before messing with the clients array! */
	gint numclients; /* Length of the previous array, i.e. the maximum amount of concurrent clients */
	void (*response_handler)( struct Request * ); /* A function that build the response */
	gboolean (*authentication_handler)( struct Request * ); /* A function that checks if we're authorized
															   It must itself set correct return codes and
															   headers. It returns authorized or not just as
															   a convenience. If not set it could be done in
															   the response handler if you make it so. */
	struct TagItem *taglist; /* The left-overs from the taglist given to server_new */
	gchar *server_hostname; /* The servers hostname */
	gpointer user_data; /* A generic user_data. Can be anything you want */
};

struct Request
{
	pthread_t thread; /* This thread */
	gint slotno; /* The index in Server->clients this Request is ranging from zero */
	gboolean disconnect; /* Set to true when we should disconnect */
	struct Server *server; /* A pointer to the server that spawned this client */

	gchar *requestline; /* The first line of the request */
	gint method; /* Request method */
	gchar *path; /* The Request URI (just the path) */
	GHashTable *get_entries; /* A parsed and split query_string. A bunch of GStrings */
	GHashTable *headers; /* All the headers */
	gchar *postdata; /* The POST body */
	GHashTable *post_entries; /* If the content-type is application/x-www-form-urlencoded
							     the post body will be decoded here just like get_entries */
	gint postlength; /* The Content-Length of the POST body */
	gint socket; /* The socket that's used */
	gint http_major, http_minor; /* HTTP/x.x */
	gchar *auth_user, *auth_pass; /* User and password from Authorization */
	GString *response; /* The response sent back */
	GList *response_headers; /* Extra headers sent back */
	GString *content_type; /* The Content-Type */
	gint ret_code; /* HTTP error code to return.
					  200 = OK
					  404 = Not found etc */
	gchar *server_hostname; /* Either is a dup of server->server_hostname or what the
							   client has said in it's Host: request header */
	gint server_port; /* If a port is specified in Host: it's put here. If not
					     what it sais in struct Server */
	gchar *clientaddr; /* The address of the client */
};

#endif
