/*
    XMMSd - A web control plug-in for XMMS
    Copyright (C) 2002  Fredrik Rambris <boost@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* util.c - Utility functions for decoding strings and stripping whitespace */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>

char x2c(char *what) {
    register char digit;

    digit = (what[0] >= 'A' ? ((what[0] & 0xdf) - 'A')+10 : (what[0] - '0'));
    digit *= 16;
    digit += (what[1] >= 'A' ? ((what[1] & 0xdf) - 'A')+10 : (what[1] - '0'));
    return digit;
}

int unescape_url(char *url) {
    register int x,y;

    for(x=0,y=0;url[y];++x,++y) {
        if((url[x] = url[y]) == '%') {
            url[x] = x2c(&url[y+1]);
            y+=2;
        }
    }
    url[x] = '\0';
	return x; /* This should be the length of the string to allow NULLs etc in the string */
}

void plustospace(char *str) {
    register int x;

    for(x=0;str[x];x++) if(str[x] == '+') str[x] = ' ';
}

/* Removes all whitespace from the beginning and end of str */
char *stripws( char *str )
{
	int start, end;
	if( !str ) return( NULL );
	for( start=0 ; str[start]==' ' || str[start]=='\n' || str[start]=='\r' ; start++ );
	if( str[start]=='\0' )
	{
		str[0]='\0';
		return( str );
	}
	for( end=strlen( str ) ; ((str[end]==' ') || (str[end]=='\0') || (str[end]=='\n') || (str[end]=='\r')) && end>-1 ; end-- ) str[end]='\0';
	if( end<1 ) return( str );
	memmove( str, str+start, end-start+1 );
	str[ end-start+1 ]='\0';
	return( str );
}

/* Replaces all occurances of needle in haystack with str */
gchar *str_replace( gchar *needle, gchar *str, gchar *haystack )
{
	GString *work;
	gchar *ptr;
	gint offset=0, needle_len=strlen( needle ), str_len=strlen( str );
	work=g_string_new( haystack );
	while( ( ptr=(gchar *)strstr( &work->str[offset], needle ) ) )
	{
		offset=GPOINTER_TO_INT(ptr)-GPOINTER_TO_INT(work->str);
		g_string_erase( work, offset, needle_len );
		g_string_insert( work, offset, str );
		offset+=str_len;
	}
	g_free( haystack );
	haystack=g_strdup( work->str );
	g_string_free( work, TRUE );
	return( haystack );
}

