/*
    XMMSd - A web control plug-in for XMMS
    Copyright (C) 2002  Fredrik Rambris <boost@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* xmmsd.c - Code for starting the server in stand-alone mode */

#include <config.h>

#define _GNU_SOURCE
#include <getopt.h>

#include "types.h"
#include "server.h"
#include "response.h"

#include <signal.h>

struct Server *serv=NULL;
gint running=1;
gint port=8335;
gchar *bind_to=NULL;
gint session=0;

void sighandler( int signal )
{
	running=0;
}

struct option long_options[] =
{
	{"help", 0, NULL, 'h'},
	{"session", 1, NULL, 's'},
	{"bind", 1, NULL, 'b'},
	{"port", 1, NULL, 'p'},
	{0,0,0,0}
};

void display_usage(void)
{
	printf( "Usage: xmmsd [options] ...\n\n"
			  "Options:\n"
			  "--------\n");
	printf( "\n-h, --help	");
	printf( "Display this text and exit.");

	printf( "\n-s, --session	");
	printf( "Select XMMS session (Default: 0)");

	printf( "\n-b, --bind	");
	printf( "IP-address of a local interface on which to allow incomming\n		connections or * (asterisk) for all interfaces.\n		(Default: 127.0.0.1)\n");

	printf( "\n-p, --port	");
	printf( "Port where clients connect to (Default: 8335)\n\n");
	exit(0);
}
void parse_cmd_line( int argc, char **argv )
{
	char c;
	while( ( c=getopt_long( argc, argv, "hs:b:p:", long_options, NULL ) ) != -1 )
	{
		switch( c )
		{
			case 'h':
				display_usage();
				break;

			case 's':
				session=atoi( optarg );
				break;

			case 'b':
				if( bind_to ) g_free( bind_to );
				bind_to=g_strdup( optarg );
				break;

			case 'p':
				port=atoi( optarg );
				break;
		}
	}
}

int main( int argc, char **argv )
{
	printf( "xmmsd-standalone v%s\n", VERSION );
	bind_to=g_strdup("127.0.0.1");
	parse_cmd_line( argc, argv );
	serv=server_new(
			SERVER_RESPONSE, GPOINTER_TO_INT( handle_http_response ),
			SERVER_BIND, bind_to,
			SERVER_PORT, port,
			XMMSD_SESSION, session,
			TAG_DONE );
	signal( SIGINT, sighandler );
	while( running ) usleep(250000);
	fprintf( stderr, "Shutting down\n" );
	server_free( serv );
	if( bind_to ) g_free( bind_to );
	bind_to=NULL;
	return( 0 );
}
